// created on 10/02/2005 at 18:56
using System;
using System.Xml;
using System.Collections;
using System.IO;
using Gtk;
public class CheezeSaver {
	public static string Serialize (Gtk.TextBuffer buffer) {
		return Serialize (buffer, buffer.StartIter, buffer.EndIter);
	}

	public static string Serialize (Gtk.TextBuffer buffer,
					Gtk.TextIter   start,
					Gtk.TextIter   end) {
		StringWriter stream = new StringWriter ();
		XmlTextWriter xml = new XmlTextWriter (stream);

		Serialize (buffer,start,end,xml);

		xml.Close ();
		return stream.ToString ();

	}

	public static void Serialize (Gtk.TextBuffer buffer,
				      Gtk.TextIter   start,
				      Gtk.TextIter   end,
				      XmlTextWriter xml) {
		Stack tag_stack = new Stack();
		Stack replay_stack = new Stack();

		Gtk.TextIter iter = start;
		Gtk.TextIter next_iter = start;
		next_iter.ForwardChar();

		xml.WriteStartElement(null,"cheezepad-content", null);
		xml.WriteAttributeString ("version", "0.0");

		foreach( Gtk.TextTag start_tag in start.Tags) {
			if(!start.TogglesTag (start_tag)){
				tag_stack.Push (start_tag);
				xml.WriteStartElement (null, start_tag.Name, null);
			}
		}

		while (!iter.Equal (end) && iter.Char != null) {
			foreach (Gtk.TextTag tag in iter.Tags) {
				if (iter.BeginsTag (tag)){
					tag_stack.Push (tag);
					xml.WriteStartElement (null, tag.Name, null);
				}
			}
			xml.WriteString (iter.Char);

			foreach(Gtk.TextTag tag in iter.Tags){
				if (TagEndsHere (tag, iter, next_iter)){
					while(tag_stack.Count > 0) {
						Gtk.TextTag existing_tag = 
							(Gtk.TextTag) tag_stack.Pop();

						if(!TagEndsHere (existing_tag,iter,next_iter))
							replay_stack.Push (existing_tag);
						xml.WriteEndElement();
					}
					
					while (replay_stack.Count > 0) {
						Gtk.TextTag replay_tag = (Gtk.TextTag) replay_stack.Pop();
						tag_stack.Push (replay_tag);
						xml.WriteStartElement (null,
								       replay_tag.Name,
								       null);
					}
				}
			}
			iter.ForwardChar();
			next_iter.ForwardChar();

		}

		while (tag_stack.Count >0) {
			tag_stack.Pop();
			xml.WriteEndElement();
		}

		xml.WriteEndElement();
	}

	static bool TagEndsHere (Gtk.TextTag tag, Gtk.TextIter iter, Gtk.TextIter next_iter){
		return (iter.HasTag(tag) && !next_iter.HasTag(tag)) || next_iter.IsEnd;
	}

	class TagStart {
		public int Start;
		public string TagName;
	}

	public static void Deserialize (Gtk.TextBuffer buffer, string content) {
		Deserialize (buffer, buffer.StartIter, content);
	}

	public static void Deserialize (Gtk.TextBuffer buffer,
					Gtk.TextIter start,
					string content){
		StringReader reader = new StringReader (content);
		XmlTextReader xml = new XmlTextReader (reader);
		xml.Namespaces = false;

		Deserialize (buffer, buffer.StartIter, xml);
	}

	public static void Deserialize (Gtk.TextBuffer buffer,
					Gtk.TextIter start,
					XmlTextReader xml) {
		int offset = start.Offset;
		Stack stack = new Stack ();
		TagStart tag_start;

		while (xml.Read ()) {
			switch (xml.NodeType) {
			case XmlNodeType.Element:
				if (xml.Name == "cheezepad-content")
					break;
				string name = xml.Name;
				
				//The following regexes determine for us whether or not
				//we are looking at a color tag, since they need to be
				//calculated and created as we find them.  All other style
				//tags already exist (bold etc...) so we can just index
				//this allows us to have arbitrary colorization (up to 
				//the max colors of a Gdk.Color obj.)
				System.Text.RegularExpressions.Regex isfg = new 
				System.Text.RegularExpressions.Regex("(Foregroundrgb)");
				System.Text.RegularExpressions.Regex isbg = new 
				System.Text.RegularExpressions.Regex("(Backgroundrgb)");
				
				if(isfg.IsMatch(name))
					deserialize_rgb_color_tag(name,true,buffer);
				else if(isbg.IsMatch(name))
					deserialize_rgb_color_tag(name,false,buffer);
					
				tag_start = new TagStart ();
				tag_start.Start = offset;
				tag_start.TagName = xml.Name;
				stack.Push (tag_start);
				break;
			case XmlNodeType.Text:
			case XmlNodeType.Whitespace:
			case XmlNodeType.SignificantWhitespace:
				Gtk.TextIter insert_at = buffer.GetIterAtOffset (offset);
				buffer.Insert (ref insert_at, xml.Value);

				offset += xml.Value.Length;
				break;
			case XmlNodeType.EndElement:
				if (xml.Name == "cheezepad-content")
					break;

				tag_start = (TagStart) stack.Pop ();

				Gtk.TextIter apply_start, apply_end;
				apply_start = buffer.GetIterAtOffset (tag_start.Start);
				apply_end = buffer.GetIterAtOffset (offset);

				buffer.ApplyTag (tag_start.TagName, apply_start, apply_end);
				break;
			default:
				break;
			}
		}
	}
	
	public static void deserialize_rgb_color_tag(string xmlNameValue,bool isFg,Gtk.TextBuffer buffer){
		System.Text.RegularExpressions.Regex num = new 
		System.Text.RegularExpressions.Regex("(-)");
		//if this tag is a Foregroundrgb tag we need to 
		//come up with a style to match it
		string color;
		if(isFg)
			color = xmlNameValue.Replace("Foregroundrgb:","");
		else
			color = xmlNameValue.Replace("Backgroundrgb:","");
			
		int i = 0;
		byte[] rgb = new byte[]{0,0,0};
		foreach(string c in num.Split(color)){
			if(!num.IsMatch(c) && c.Length > 0){
				rgb[i++]=byte.Parse(c);							
			}
		}
		if(buffer.TagTable.Lookup(xmlNameValue)==null){
			Gtk.TextTag colortag = new Gtk.TextTag(xmlNameValue);
			if(isFg){
				colortag.ForegroundGdk = 
				new Gdk.Color((byte)rgb[0],
							  (byte)rgb[1],
							  (byte)rgb[2]);
			}else{
			    colortag.BackgroundGdk = 
				new Gdk.Color((byte)rgb[0],
							  (byte)rgb[1],
							  (byte)rgb[2]);
			}
			buffer.TagTable.Add(colortag);
		}
	}
		
}

					